# pangadfs/pangadfs/populate.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from typing import Dict

import numpy as np

from pangadfs.base import PopulateBase
from pangadfs.misc import multidimensional_shifting


class PopulateDefault(PopulateBase):

    def populate(self,
                 *, 
                 pospool, 
                 posmap: Dict[str, int], 
                 population_size: int, 
                 probcol: str='prob',
                 **kwargs) -> np.ndarray:
        """Creates individuals in population
        
        Args:
            pospool (Dict[str, DataFrame]): pool split into positions
            posmap (Dict[str, int]): positions & accompanying roster slots
            population_size (int): number of individuals to create
            probcol (str): the dataframe column with probabilities
            **kwargs: keyword arguments

        Returns:
            ndarray of size (population_size, sum(posmap.values()))

        """
        pos_samples = {
            pos: multidimensional_shifting(pospool[pos].index, population_size, n, pospool[pos][probcol])
            for pos, n in posmap.items()
        }

        # concatenate positions into single row
        pop = np.concatenate([pos_samples[pos] for pos in posmap if pos != 'FLEX'], axis=1)

        # find non-duplicate FLEX and aggregate with other positions
        # https://stackoverflow.com/questions/65473095
        # https://stackoverflow.com/questions/54155844/
        dups = (pos_samples['FLEX'][..., None] == pop[:, None, :]).any(-1)
        return np.column_stack((pop, pos_samples['FLEX'][np.invert(dups).cumsum(axis=1).cumsum(axis=1) == 1]))
 


# pangadfs/pangadfs/select.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import numpy as np
from pangadfs.base import SelectBase


class SelectDefault(SelectBase):

    def _fittest(self,
                 *,
                 population: np.ndarray, 
                 population_fitness: np.ndarray,
                 n: int,
                 **kwargs) -> np.ndarray:
        """n-fittest selection of individuals in population. 
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            n (int): total number of individuals to select
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        # argpartition produces an unsorted rearrangement
        # use negative parameter and slice to get top n values
        return population[np.argpartition(population_fitness, -n, axis=0)][-n:]

    def _rank(self, 
               *, 
               population: np.ndarray, 
               population_fitness: np.ndarray,
               n: int,
               **kwargs) -> np.ndarray:
        """Rank selection of individuals in population. 
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            n (int): total number of individuals to select
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        elements = len(population)
        temp = population_fitness.argsort()
        ranks = np.empty_like(temp)
        ranks[temp] = np.arange(1, elements + 1)
        weights = ranks / ranks.sum()
        return population[np.random.choice(elements, size=n, replace=False, p=weights)]

    def _roulette_wheel(self, 
                        *, 
                        population: np.ndarray, 
                        population_fitness: np.ndarray,
                        n: int,
                        **kwargs) -> np.ndarray:
        """Select individuals in population using stochastic universal sampling, which samples without replacement.
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            n (int): total number of individuals to select
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        # alternative implementation if np.random.choice changes
        # fitness_cumsum = fitness.cumsum()   # the "roulette wheel"
        # fitness_sum = fitness_cumsum[-1]    # sum of all fitness values (size of the wheel)
        # sampled_values = np.random.random(size) * fitness_sum
        # For each sampled value, get the corresponding roulette wheel slot
        # selected = np.searchsorted(fitness_cumsum, sampled_values)
        # return selected
        weights = population_fitness / population_fitness.sum()
        return population[np.random.choice(len(population_fitness), size=n, replace=True, p=weights)]       

    def _scaled(self, 
               *, 
               population: np.ndarray, 
               population_fitness: np.ndarray,
               n: int,
               **kwargs) -> np.ndarray:
        """Select individuals in population using scaled fitness.
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            n (int): total number of individuals to select
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        scaled = (population_fitness - population_fitness.min()) / (population_fitness.max() - population_fitness.min())
        weights = scaled / scaled.sum()
        return population[np.random.choice(len(population), size=n, replace=False, p=weights)]

    def _sus(self, 
             *, 
             population: np.ndarray, 
             population_fitness: np.ndarray,
             n: int,
             **kwargs) -> np.ndarray:
        """Select individuals in population using stochastic universal sampling, which samples without replacement.
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            n (int): total number of individuals to select
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        # cumsum creates the roulette wheel
        # is order-insensitive: if 1st element largest, starts there
        fitness_cumsum = population_fitness.cumsum()
        fitness_sum = fitness_cumsum[-1]

        # so if total is 100 and n is 10
        # 1st point is 10 starting point is first element > 10
        step = fitness_sum / n
        start = np.random.random() * step

        # selectors are the evenly-spaced points on wheel
        selectors = np.arange(start, fitness_sum, step)

        # Find indices where selectors should be inserted to maintain order.
        # so if fitness is 1, 2, 3, 4, 5 and selectors 1.2 and 3.1
        # np.searchsorted would be [1, 3]
        return population[np.searchsorted(fitness_cumsum, selectors)]
     
    def _tournament(self, 
                    *, 
                    population: np.ndarray, 
                    population_fitness: np.ndarray,
                    tournament_size: int = 2,
                    **kwargs) -> np.ndarray:
        """Tournament selection of individuals in population. 
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            tournament_size (int): number of individuals to compete against each other
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        max_idx = len(population_fitness) // tournament_size * tournament_size
        tournament_fitness = population_fitness[:max_idx].reshape(-1, tournament_size)
        winners = np.argmax(tournament_fitness, axis=1) + np.arange(max_idx, step=tournament_size)
        return population[winners]

    def select(self, 
               *, 
               population: np.ndarray, 
               population_fitness: np.ndarray,
               n: int,
               method: str = 'roulette',
               **kwargs) -> np.ndarray:
        """Select individuals in population using specified method.
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            population_fitness (np.ndarray): the population fitness. Is a 1D array same length as population.
            n (int): total number of individuals to select
            method (str): 'roulette', 'su', 'scaled'; default 'roulette'
            **kwargs: keyword arguments for plugins

        Returns:
            np.ndarray: selected population

        """
        dispatch = {
            'roulette': self._roulette_wheel,
            'sus': self._sus,
            'rank': self._rank,
            'tournament': self._tournament,
            'scaled': self._scaled,
            'fittest': self._fittest
        }        

        params = {
            'population': population,
            'population_fitness': population_fitness,
            'n': n
        }

        return dispatch.get(method, self._roulette_wheel)(**params, **kwargs)



# pangadfs/pangadfs/fitness.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import numpy as np

from pangadfs.base import FitnessBase


class FitnessDefault(FitnessBase):

    def fitness(self,
                *, 
                population: np.ndarray, 
                points: np.ndarray,
                **kwargs):
        """Assesses population fitness using supplied mapping
        
        Args:
            population (np.ndarray): the population to assess fitness
            points (np.ndarray): 1D array of projected points in same order as pool
            **kwargs: Arbitrary keyword arguments

        Returns:
            np.ndarray: 1D array of float

        """
        return np.sum(points[population], axis=1)





# panggadfs/app/configapp/gasettings.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from dataclasses import dataclass
import logging
from pathlib import Path
from typing import Any, Dict, Iterable

import dacite


logging.getLogger(__name__).addHandler(logging.NullHandler())


@dataclass
class GASettings:
    csvpth: str = 'pool.csv'
    crossover_method: str = 'uniform'
    elite_divisor: int = 5
    elite_method: str = 'fittest'
    mutation_rate: float = 0.05
    n_generations: int = 20
    points_column: str = 'proj'
    population_size: int = 5000
    position_column: str = 'pos'
    salary_column: str = 'salary'
    select_method: str = 'roulette'
    stop_criteria: int = 10
    verbose: bool = True


@dataclass
class PluginSettings:
    driver_managers: Dict[str, Any] = None
    extension_managers: Dict[str, Any] = None


@dataclass
class SiteSettings:
    lineup_size: int
    posfilter: Dict[str, int]
    posmap: Dict[str, int]
    salary_cap: int
    flex_positions: Iterable = ('RB', 'WR', 'TE')


@dataclass
class AppSettings:
    ga_settings: GASettings
    site_settings: SiteSettings
    plugin_settings: PluginSettings


def ctx_from_dict(raw_cfg):
    """Creates config object from dictionary"""

    # create and validate the Configuration object
    return dacite.from_dict(
        data_class=AppSettings, 
        data=raw_cfg
    )


# pangadfs/pangadfs/mutate.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import numpy as np

from pangadfs.base import MutateBase


class MutateDefault(MutateBase):

    def mutate(self, *, population: np.ndarray, mutation_rate: float = .05) -> np.ndarray:
        """Mutates individuals in population
        
        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            mutation_rate (float): decimal value from 0 to 1, default .05

        Returns:
            np.ndarray: same shape as population

        """
        # mutate is ndarray of same shape of population of dtype bool
        # if mutation_rate is .05, then 1 out of 20 values should be True
        # where mutate is true, swap randomly-selected player into population
        # ensures swap comes from same lineup slot, but does not prevent duplicates from other slots
        # so lineup positional allocation will stay valid, but duplicates are possible
        mutate = (
            np.random.binomial(n=1, p=mutation_rate, size=population.size)
            .reshape(population.shape)
            .astype(bool)
        )
        swap = population[np.random.choice(len(population), size=len(population), replace=False)]
        return np.where(mutate, swap, population)



# panggadfs/app/configapp/optimizer.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import logging
from typing import Any, Dict, Tuple

import numpy as np

from pangadfs.ga import GeneticAlgorithm
from gasettings import AppSetting


class Optimizer:

    def __init__(self, 
                 ctx: AppSetting, 
                 driver_managers: Dict[str, Any], 
                 extension_managers: Dict[str, Any]):
        """Creates GeneticAlgorithm instance

        Args:
            ctx (AppSetting): context dataclass with all relevant settings
            driver_managers (Dict[str, Any]): driver managers
            extension_managers (Dict[str, Any]): extension managers

        Returns:
            Optimizer: the optimizer instance

        """
        logging.getLogger(__name__).addHandler(logging.NullHandler())

        # app settings
        self.ctx = ctx

        # create ga
        self.ga = GeneticAlgorithm(
            driver_managers=driver_managers, 
            extension_managers=extension_managers
        )

    def optimize(self, verbose: bool = True, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """Default routine for optimizing lineup
        
        Args:
            verbose (bool): controls logging of optimizer progress
            **kwargs: Keyword arguments for plugins (other than default)
        
        Returns:
            Tuple[np.ndarray, np.ndarray]: population and population fitness arrays                
        
        """
        pop_size = self.ctx.ga_settings.population_size
        pool = self.ga.pool(csvpth=self.ctx.ga_settings.csvpth)
        cmap = {'points': self.ctx.ga_settings.points_column,
			    'position': self.ctx.ga_settings.position_column,
			    'salary': self.ctx.ga_settings.salary_column}
        posfilter = self.ctx.site_settings.posfilter
        pospool = self.ga.pospool(
            pool=pool, 
            posfilter=posfilter, 
            column_mapping=cmap, 
            flex_positions=self.ctx.site_settings.flex_positions
        )

        # create dict of index and stat value
        # this will allow easy lookup later on
        cmap = {'points': self.ctx.ga_settings.points_column,
                'salary': self.ctx.ga_settings.salary_column}
        points = pool[cmap['points']].values
        salaries = pool[cmap['salary']].values
        
        # CREATE INITIAL POPULATION
        initial_population = self.ga.populate(
            pospool=pospool, 
            posmap=self.ctx.site_settings.posmap, 
            population_size=pop_size
        )

        # apply validators
        initial_population = self.ga.validate(
            population=initial_population, 
            salaries=salaries,
            salary_cap=self.ctx.site_settings.salary_cap
        )

        population_fitness = self.ga.fitness(
            population=initial_population, 
            points=points
        )

        # set overall_max based on initial population
        omidx = population_fitness.argmax()
        best_fitness = population_fitness[omidx]
        best_lineup = initial_population[omidx]

        # CREATE NEW GENERATIONS
        n_unimproved = 0
        for i in range(1, self.ctx.ga_settings.n_generations + 1):
            if n_unimproved == self.ctx.ga_settings.stop_criteria:
                break
            if i == 1:
                population = initial_population

            if verbose:
                logging.info(f'Starting generation {i}')
                logging.info(f'Best lineup score {best_fitness}')

            elite = self.ga.select(
                population=population, 
                population_fitness=population_fitness, 
                n=len(population) // self.ctx.ga_settings.elite_divisor,
                method=self.ctx.ga_settings.elite_method
            )

            selected = self.ga.select(
                population=population, 
                population_fitness=population_fitness, 
                n=len(population),
                method=self.ctx.ga_settings.select_method
            )

            # uniform crossover
            crossed_over = self.ga.crossover(
                population=selected, 
                method=self.ctx.ga_settings.crossover_method
            )

            # reduce mutation over generations
            mutation_rate = max(self.ctx.ga_settings.mutation_rate, n_unimproved / 50)
            mutated = self.ga.mutate(population=crossed_over, mutation_rate=mutation_rate)

            population = self.ga.validate(
                population=np.vstack((elite, mutated)), 
                salaries=salaries, 
                salary_cap=self.ctx.site_settings.salary_cap
            )
            
            population_fitness = self.ga.fitness(population=population, points=points)
            omidx = population_fitness.argmax()
            generation_max = population_fitness[omidx]
        
            if generation_max > best_fitness:
                logging.info(f'Lineup improved to {generation_max}')
                best_fitness = generation_max
                best_lineup = population[omidx]
                n_unimproved = 0
            else:
                n_unimproved += 1
                logging.info(f'Lineup unimproved {n_unimproved} times')

        # FINALIZE RESULTS
        if verbose:
            print(pool.loc[best_lineup, :])
            print(f'Lineup score: {best_fitness}')
        
        return population, population_fitness



# pangadfs/pangadfs/ga.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import logging
from pathlib import Path
from typing import Dict, Iterable

import numpy as np
import pandas as pd
from stevedore.driver import DriverManager
from stevedore.named import NamedExtensionManager


class GeneticAlgorithm:
    """Handles coordination of genetic algorithm plugins"""

    PLUGIN_NAMESPACES = (
       'pool', 'pospool', 'populate', 'fitness', 
       'select', 'crossover', 'mutate', 'validate'
    )

    VALIDATE_PLUGINS = ('validate_salary', 'validate_duplicates')

    def __init__(self, 
                 driver_managers: Dict[str, DriverManager] = None, 
                 extension_managers: Dict[str, NamedExtensionManager] = None,
                 use_defaults: bool = False):
        """Creates GeneticAlgorithm instance

        Args:
            driver_managers (dict): key is namespace, value is DriverManager
            extension_managers (dict): key is namespace, value is NamedExtensionManager
            use_defaults (bool): use default plugins

        Returns:
            GeneticAlgorithm: the GA instance

        """
        logging.getLogger(__name__).addHandler(logging.NullHandler())

        # add driver/extension managers
        self.driver_managers = driver_managers if driver_managers else {}
        self.extension_managers = extension_managers if extension_managers else {}

        # if use_defaults, then load default plugin(s) for missing namespaces
        if use_defaults:
            self._load_plugins()

    def _load_plugins(self):
        """Loads default plugins for any namespace that doesn't have a plugin"""
        for ns in self.PLUGIN_NAMESPACES:
            if ns not in self.driver_managers and ns not in self.extension_managers:
                if ns == 'validate':
	                self.extension_managers[ns] = NamedExtensionManager(
                        namespace='pangadfs.validate', 
                        names=self.VALIDATE_PLUGINS, 
                        invoke_on_load=True, 
                        name_order=True
                    )
                else:
                    mgr = DriverManager(
                        namespace=f'pangadfs.{ns}', 
                        name=f'{ns}_default', 
                        invoke_on_load=True
                    )
                    self.driver_managers[ns] = mgr

    def crossover(self,
                  *,
                  population: np.ndarray = None,
                  agg: bool = None,
                  **kwargs) -> np.ndarray:
        """Crossover operation to generate new population

        Args:
            population (np.ndarray): the population to cross over, is 2D array
            agg (bool): if True, then aggregate multiple crossovers, otherwise is sequential.          
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            np.ndarray: the crossed-over population

        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        # if there is a driver, then use it and run once
        if mgr := self.driver_managers.get('crossover'):
            return mgr.driver.crossover(**params, **kwargs)

        # if agg=True, then aggregate crossed over populations
        if kwargs.get('agg'):
            pops = []
            for ext in self.extension_managers['crossover'].extensions:
                try:
                    pops.append(ext.obj.mutate(**params, **kwargs))
                except:
                    continue
            return np.aggregate(pops)

        # otherwise, run crossover for each plugin
        # after first time, crosses over prior crossed-over population
        population = params['population']
        for ext in self.extension_managers['crossover'].extensions:
            try:
                params['population'] = population
                population = ext.obj.crossover(**kwargs)
            except:
                continue
        return population

    def fitness(self, 
                *,
                population: np.ndarray = None, 
                points: np.ndarray = None, 
                **kwargs) -> np.ndarray:
        """Measures fitness of population

        Args:
            population (np.ndarray): the population to cross over, is 2D array
            points (np.ndarray): the fitness of the population to crossover, is 1D array
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            np.ndarray: population fitness as 1D array of float

        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        # if there is a driver, then use it and run once
        if mgr := self.driver_managers.get('fitness'):
            return mgr.driver.fitness(**params, **kwargs)

        # otherwise, return fitness for first valid plugin
        for ext in self.extension_managers['fitness'].extensions:
            try:
                return ext.obj.fitness(**params, **kwargs)
            except:
                continue

    def mutate(self, 
               *,
               population: np.ndarray = None,
               mutation_rate: float = None,
               **kwargs) -> np.ndarray:
        """Mutates population at frequency of mutation_rate

        Args:
            population (np.ndarray): the population to mutate. Shape is n_individuals x n_chromosomes.
            mutation_rate (float): decimal value from 0 to 1, default .05
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            np.ndarray: same shape and dtype as population

        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        # if there is a driver, then use it and run once
        if mgr := self.driver_managers.get('mutate'):
            return mgr.driver.mutate(**params, **kwargs)

        # otherwise, mutate with first valid plugin
        for ext in self.extension_managers['mutate'].extensions:
            try:
                return ext.obj.mutate(**params, **kwargs)
            except:
                continue

    def pool(self, *, csvpth: Path = None, **kwargs) -> pd.DataFrame:
        """Creates pool of players.

        Args:
            csvpth (Path): the path of the datafile
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            pd.DataFrame: initial pool of players
        
        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        if mgr := self.driver_managers.get('pool'):
            return mgr.driver.pool(**params, **kwargs)   
        for ext in self.extension_managers['pool'].extensions:
            try:
                return ext.obj.pool(**params, **kwargs)  
            except:
                logging.error(f'Could not load {ext}')

    def populate(self,
                 *,
                 pospool: Dict[str, pd.DataFrame] = None,
                 posmap: Dict[str, int] = None,
                 population_size: int = None,
                 probcol: str = 'prob',
                 agg: bool = 'False',
                 **kwargs) -> np.ndarray:
        """Creates initial population of specified size
        
        Args:
            pospool (Dict[str, pd.DataFrame]): pool segmented by position
            posmap (Dict[str, int]): positions & accompanying roster slots
            population_size (int): number of individuals to create
            probcol (str): the dataframe column with probabilities, default 'probs'
            agg (bool): default False. Aggregate multiple crossovers if True.
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            np.ndarray: the population

        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        agg = params.pop('agg')
        kwargs = params.pop('kwargs')

        # if there is a driver, then use it and run once
        if mgr := self.driver_managers.get('populate'):
            return mgr.driver.populate(**params, **kwargs)

        # if agg=True, then aggregate populations
        if agg:
            pops = []
            for ext in self.extension_managers['populate'].extensions:
                try:
                    pops.append(ext.obj.populate(**params, **kwargs))  
                except:
                    continue
            return np.concatenate(pops)

        # otherwise, run populate using first valid plugin
        for ext in self.extension_managers['crossover'].extensions:
            try:
                return ext.obj.populate(**kwargs)
            except:
                continue

    def pospool(self, 
                *,
                pool: pd.DataFrame = None,
                posfilter: Dict[str, int] = None,
                column_mapping: Dict[str, str] = None,
                flex_positions: Iterable[str] = None,
                **kwargs) -> Dict[str, pd.DataFrame]:
        """Divides pool into positional buckets
        
        Args:   
            pool (pd.DataFrame):
            posfilter (Dict[str, int]): position name and points threshold
            column_mapping (Dict[str, str]): column names for player, position, salary, projection
            flex_positions (Iterable[str]): e.g. (WR, RB, TE)
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            Dict[str, pd.DataFrame] where keys == posfilter.keys

        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        # if there is a driver, then use it and run once
        # otherwise, run pospool using first valid plugin
        if mgr := self.driver_managers.get('pospool'): 
            return mgr.driver.pospool(**params, **kwargs)
        for ext in self.extension_managers['pospool'].extensions:
            try:
                return ext.obj.pospool(**params, **kwargs)  
            except:
                continue
        
    def select(self,
               *, 
               population: np.ndarray = None, 
               population_fitness: np.ndarray = None,
               n: int = None,
               method: str = 'fittest',
               **kwargs) -> np.ndarray:
        """Selects/filters population

        Args:
            population (np.ndarray): the population to cross over, is 2D array
            population_fitness (np.ndarray): 1D array of float
            n (int): number of individuals to select
            method (str): the selection method, default roulette
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            np.ndarray: population fitness as 1D array of float

        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        # if there is a driver, then use it and run once
        if mgr := self.driver_managers.get('select'):
            return mgr.driver.select(**params, **kwargs)

        # otherwise, return select for first valid plugin
        for ext in self.extension_managers['select'].extensions:
            try:
                return ext.obj.select(**params, **kwargs)
            except:
                continue

    def validate(self,
                 *,
                 population: np.ndarray = None, 
                 salaries: np.ndarray = None, 
                 **kwargs) -> np.ndarray:
        """Validate lineup according to validate plugin criteria
        
        Args:
            population (np.ndarray): the population to validate.
            salaries (np.ndarray): the population salaries
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            np.ndarray: same width and dtype as population. Likely less rows due to exclusions.
            
        """
        # combine keyword arguments with **kwargs
        params = locals().copy()
        params.pop('self', None)
        kwargs = params.pop('kwargs')

        if mgr := self.driver_managers.get('validate'):
            return mgr.driver.validate(**params, **kwargs)
        population = params['population']
        for ext in self.extension_managers['validate'].extensions:
            params['population'] = population
            population = ext.obj.validate(**params, **kwargs)
        return population



# pangadfs/pangadfs/penalty.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import numpy as np


def diversity_penalty(population: np.ndarray = None) -> np.ndarray:
    """Calculates diversity penalty for overlapping lineups
    
    Args:
        population (np.ndarray): the population

    Returns:
        np.ndarray: 1D array of float

    """
    uniques = np.unique(population)
    a = (population[..., None] == uniques).sum(1)
    out = np.einsum('ij,kj->ik', a, a)
    diversity = np.sum(out, axis=1) / population.size
    return 0 - ((diversity - diversity.mean()) / diversity.std())


def ownership_penalty(ownership, base=3, boost=2) -> np.ndarray:
    """Returns 1D array of penalties that are inverse to projected ownership
    
    Args:
        ownership (np.ndarray): 1D array of ownership
        base (int): the logarithm base, default 3
        boost (int): the constant to boost low-owned players
        
    Returns:
        np.ndarray: 1D array of penalties
        
    """
    return 0 - np.log(ownership) / np.log(base) + boost
    



# pangadfs/pangadfs/pool.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from typing import Dict
from pathlib import Path

import pandas as pd

from pangadfs.base import PoolBase


class PoolDefault(PoolBase):

    def pool(self, 
             *, 
             csvpth: Path,
             thresh: int = 4,
             **kwargs
             ) -> pd.DataFrame:
        """Creates initial pool
        
        Args:
            csvpth (Path): path for csv file
            thresh (int): global filter on low-scoring players
            **kwargs: keyword arguments for plugins
            
        Returns:
            DataFrame with columns:
            player, team, pos, salary, proj

        """
        df = pd.read_csv(csvpth)
        assert set(['player', 'team', 'pos', 'salary', 'proj']).issubset(df.columns)
        df = df.loc[df.proj >= thresh, :]
        return df.sort_values(['pos']).reset_index(drop=True)




# pangadfs/app/basicapp/app.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import logging

from pangadfs.ga import GeneticAlgorithm
from pathlib import Path

import numpy as np
from stevedore.driver import DriverManager
from stevedore.named import NamedExtensionManager


def run():
	"""Example application using pangadfs"""
	logging.basicConfig(level=logging.INFO)

	ctx = {
		'ga_settings': {
			'crossover_method': 'uniform',
			'csvpth': Path(__file__).parent.parent / 'appdata' / 'pool.csv',
			'elite_divisor': 5,
			'elite_method': 'fittest',
			'mutation_rate': .05,
			'n_generations': 20,
			'points_column': 'proj',
			'population_size': 30000,
			'position_column': 'pos',
			'salary_column': 'salary',
			'select_method': 'roulette',
			'stop_criteria': 10,
			'verbose': True
		},

		'site_settings': {
			'flex_positions': ('RB', 'WR', 'TE'),
			'lineup_size': 9,
			'posfilter': {'QB': 14, 'RB': 8, 'WR': 8, 'TE': 5, 'DST': 4, 'FLEX': 8},
			'posmap': {'DST': 1, 'QB': 1, 'TE': 1, 'RB': 2, 'WR': 3, 'FLEX': 7},
			'salary_cap': 50000
		}
	}

	# set up driver managers
	dmgrs = {}
	emgrs = {}
	for ns in GeneticAlgorithm.PLUGIN_NAMESPACES:
		pns = f'pangadfs.{ns}'
		if ns == 'validate':
			emgrs['validate'] = NamedExtensionManager(
				namespace=pns, 
				names=['validate_salary', 'validate_duplicates'], 
				invoke_on_load=True, 
				name_order=True)
		else:
			dmgrs[ns] = DriverManager(
				namespace=pns, 
				name=f'{ns}_default', 
				invoke_on_load=True)
	
	
	# set up GeneticAlgorithm object
	ga = GeneticAlgorithm(driver_managers=dmgrs, extension_managers=emgrs)
	
	# create pool and pospool
	# pospool used to generate initial population
	# is a dict of position_name: DataFrame
	pop_size = ctx['ga_settings']['population_size']
	pool = ga.pool(csvpth=ctx['ga_settings']['csvpth'])
	cmap = {'points': ctx['ga_settings']['points_column'],
			'position': ctx['ga_settings']['position_column'],
			'salary': ctx['ga_settings']['salary_column']}
	posfilter = ctx['site_settings']['posfilter']
	flex_positions = ctx['site_settings']['flex_positions']
	pospool = ga.pospool(pool=pool, posfilter=posfilter, column_mapping=cmap, flex_positions=flex_positions)

	# create salary and points arrays
	# these match indices of pool
	cmap = {'points': ctx['ga_settings']['points_column'],
			'salary': ctx['ga_settings']['salary_column']}
	points = pool[cmap['points']].values
	salaries = pool[cmap['salary']].values
	
	# create initial population
	initial_population = ga.populate(
		pospool=pospool, 
		posmap=ctx['site_settings']['posmap'], 
		population_size=pop_size
	)

	# apply validators
	# default is to valdate duplicates and salary
	# can add other validators as desired
	initial_population = ga.validate(
		population=initial_population, 
		salaries=salaries,
		salary_cap=ctx['site_settings']['salary_cap']
	)

	# need fitness to determine best lineup
	# and also for selection when loop starts
	population_fitness = ga.fitness(
		population=initial_population, 
		points=points
	)

	# set overall_max based on initial population
	omidx = population_fitness.argmax()
	best_fitness = population_fitness[omidx]
	best_lineup = initial_population[omidx]

	# create new generations
	n_unimproved = 0
	population = initial_population.copy()

	for i in range(1, ctx['ga_settings']['n_generations'] + 1):

		# end program after n generations if not improving
		if n_unimproved == ctx['ga_settings']['stop_criteria']:
			break

		# display progress information with verbose parameter
		if ctx['ga_settings'].get('verbose'):
			logging.info(f'Starting generation {i}')
			logging.info(f'Best lineup score {best_fitness}')

		# select the population
		# here, we are holding back the fittest 20% to ensure
		# that crossover and mutation do not overwrite good individuals
		elite = ga.select(
			population=population, 
			population_fitness=population_fitness, 
			n=len(population) // ctx['ga_settings'].get('elite_divisor', 5),
			method=ctx['ga_settings'].get('elite_method', 'fittest')
		)

		selected = ga.select(
			population=population, 
			population_fitness=population_fitness, 
			n=len(population),
			method=ctx['ga_settings'].get('select_method', 'roulette')
		)

		# cross over the population
		# here, we use uniform crossover, which splits the population
		# and randomly exchanges 0 - all chromosomes
		crossed_over = ga.crossover(population=selected, method=ctx['ga_settings'].get('crossover_method', 'uniform'))

		# mutate the crossed over population (leave elite alone)
		# can use fixed rate or variable to reduce mutation over generations
		# here we use a variable rate that increases if no improvement is found
		mutation_rate = ctx['ga_settings'].get('mutation_rate', max(.05, n_unimproved / 50))
		mutated = ga.mutate(population=crossed_over, mutation_rate=mutation_rate)

		# validate the population (elite + mutated)
		population = ga.validate(
			population=np.vstack((elite, mutated)), 
			salaries=salaries, 
			salary_cap=ctx['site_settings']['salary_cap']
		)
		
		# assess fitness and get the best score
		population_fitness = ga.fitness(population=population, points=points)
		omidx = population_fitness.argmax()
		generation_max = population_fitness[omidx]
	
		# if new best score, then set n_unimproved to 0
		# and save the new best score and lineup
		# otherwise increment n_unimproved
		if generation_max > best_fitness:
			logging.info(f'Lineup improved to {generation_max}')
			best_fitness = generation_max
			best_lineup = population[omidx]
			n_unimproved = 0
		else:
			n_unimproved += 1
			logging.info(f'Lineup unimproved {n_unimproved} times')

	# FINALIZE RESULTS
	# show best score and lineup at conclusion
	# will break after n_generations or when stop_criteria reached
	print(pool.loc[best_lineup, :])
	print(f'Lineup score: {best_fitness}')
	

if __name__ == '__main__':
	run()



# gadfs/gadfs/base.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import abc
import logging


class CrossoverBase(metaclass=abc.ABCMeta):
    """Base class for crossover plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def crossover(self, *args, **kwargs):
        """Implement crossover."""


class FitnessBase(metaclass=abc.ABCMeta):
    """Base class for fitness plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def fitness(self, *args, **kwargs):
        """Implement fitness."""


class MutateBase(metaclass=abc.ABCMeta):
    """Base class for crossover plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def mutate(self, *args, **kwargs):
        """Mutates population at given mutation rate."""


class PopulateBase(metaclass=abc.ABCMeta):
    """Base class for populate plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def populate(self, *args, **kwargs):
        """Creates initial population from pool."""


class PoolBase(metaclass=abc.ABCMeta):
    """Base class for pool plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def pool(self, *args, **kwargs):
        """Implement pool."""


class PospoolBase(metaclass=abc.ABCMeta):
    """Base class for pospool plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def pospool(self, *args, **kwargs):
        """Implement pospool."""


class SelectBase(metaclass=abc.ABCMeta):
    """Base class for select plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def select(self, *args, **kwargs):
        """Implement select."""


class ValidateBase(metaclass=abc.ABCMeta):
    """Base class for validate plugins."""

    def __init__(self):
        logging.getLogger(__name__).addHandler(logging.NullHandler())

    @abc.abstractmethod
    def validate(self, *args, **kwargs):
        """Implement validate."""



# pangadfs/pangadfs/crossover.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from typing import Tuple
import numpy as np

from pangadfs.base import CrossoverBase
from pangadfs.misc import diversity, parents


class CrossoverDefault(CrossoverBase):

    def _diverse(self, *, population: np.ndarray, **kwargs) -> np.ndarray:
        """Diverse crossover of individuals in population.
        
        Args:
            population (np.ndarray): the population to crossover. Shape is n_individuals x n_chromosomes.
            **kwargs: Arbitrary keyword arguments

        Returns:
            np.ndarray: concatenation of two offspring

        """     
        # crosses over individuals with least common elements
        # step one is to get diversity matrix
        # then get indices of the minimum of each row (randomize selection among duplicates)
        diversity_matrix = diversity(population)
        cxidx = np.argmax(np.random.random(diversity_matrix.shape) * (diversity_matrix == diversity_matrix.min()), axis=1)
        choice = np.random.randint(2, size=diversity_matrix.size).reshape(diversity_matrix.shape).astype(bool)   
        return np.where(choice, diversity_matrix, cxidx)

    def _one_point(self, *, population: np.ndarray, point: int = 3, **kwargs) -> np.ndarray:
        """Crosses over individuals in population at a single point.
        
        Args:
            population (np.ndarray): the population to crossover. Shape is n_individuals x n_chromosomes.
            point (int): the crossover point, default 3
            **kwargs: Arbitrary keyword arguments
        
        Returns:
            np.ndarray: concatenation of two offspring

        """
        fathers, mothers = parents(population)
        child1 = np.hstack((fathers[:, 0:point], mothers[:, point:]))
        child2 = np.hstack((mothers[:, 0:point], fathers[:, point:]))
        return np.vstack((child1, child2))

    def _two_point(self, *, population: np.ndarray, points: Tuple[int] = (3, 7), **kwargs) -> np.ndarray:
        """Crosses over individuals in population at two points.
        
        Args:
            population (np.ndarray): the population to crossover. Shape is n_individuals x n_chromosomes.
            points (Tuple[int]): the crossover points, default (3, 7)
            **kwargs: Arbitrary keyword arguments    
    
        Returns:
            np.ndarray: concatenation of two offspring

        """
        fathers, mothers = parents(population)
        point1, point2 = points
        child1 = np.hstack((fathers[:, 0:point1], 
                            mothers[:, point1:point2],
                            fathers[:, point2:]))
        child2 = np.hstack((mothers[:, 0:point1], 
                            fathers[:, point1:point2],
                            mothers[:, point2:]))
        return np.vstack((child1, child2))

    def _uniform(self, *, population: np.ndarray, **kwargs) -> np.ndarray:
        """Uniform crossover of individuals in population.
        
        Args:
            population (np.ndarray): the population to crossover. Shape is n_individuals x n_chromosomes.
            **kwargs: Arbitrary keyword arguments

        Returns:
            np.ndarray: concatenation of two offspring

        """     
        fathers, mothers = parents(population)
        choice = np.random.randint(2, size=fathers.size).reshape(fathers.shape).astype(bool)   
        return np.vstack((np.where(choice, fathers, mothers), np.where(choice, mothers, fathers)))

    def crossover(self, *, population: np.ndarray, method: str = 'uniform', **kwargs) -> np.ndarray:
        """Crossover individuals in population.
        
        Args:
            population (np.ndarray): the population to crossover. Shape is n_individuals x n_chromosomes.
            method (str): crossover method, 'uniform', 'diverse', 'one_point', or 'two_point'.
            **kwargs: Arbitrary keyword arguments

        Returns:
            np.ndarray: concatenation of two offspring

        """     
        dispatch = {
            'uniform': self._uniform,
            'diverse': self._diverse,
            'one_point': self._one_point,
            'two_point': self._two_point
        }        

        return dispatch.get(method, self._uniform)(population=population, **kwargs)



# -*- coding: utf-8 -*-
"""
setup.py

installation script

"""

from setuptools import setup, find_packages

PACKAGE_NAME = "pangadfs"


def run():
    setup(
        name=PACKAGE_NAME,
        version="0.1",
        description="extensible pandas-based genetic algorithm for fantasy sports lineup optimization",
        author="Eric Truett",
        author_email="eric@erictruett.com",
        license="MIT",
        packages=find_packages(),
        entry_points={
          'pangadfs.crossover': ['crossover_default = pangadfs.crossover:CrossoverDefault'],
          'pangadfs.select': ['select_default = pangadfs.select:SelectDefault'],
          'pangadfs.mutate': ['mutate_default = pangadfs.mutate:MutateDefault'],
          'pangadfs.populate': ['populate_default = pangadfs.populate:PopulateDefault'],
          'pangadfs.fitness': ['fitness_default = pangadfs.fitness:FitnessDefault'],
          'pangadfs.validate': ['validate_salary = pangadfs.validate:SalaryValidate',
                                'validate_duplicates = pangadfs.validate:DuplicatesValidate'],
          'pangadfs.pool': ['pool_default = pangadfs.pool:PoolDefault'],
          'pangadfs.pospool': ['pospool_default = pangadfs.pospool:PospoolDefault'],
          'console_scripts': ['basicapp=pangadfs.app.basicapp.app:run', 
                              'configapp=pangadfs.app.configapp.app:run']
        },
        zip_safe=False,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.8',
    )


if __name__ == "__main__":
    run()



# pangadfs/pangadfs/postprocess.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from typing import Dict
import numpy as np


def exposure(population: np.ndarray = None) -> Dict[int, int]:
    """Returns dict of index: count of individuals

    Args:
        population (np.ndarray): the population

    Returns:
        Dict[int, int]: key is index, value is count of lineup

    Examples:
        >>> fittest_population = population[np.where(fitness > np.percentile(fitness, 97))]
        >>> exposure = population_exposure(fittest_population)
        >>> top_exposure = np.argpartition(np.array(list(exposure.values())), -10)[-10:]
        >>> print([round(i, 3) for i in sorted(top_exposure / len(fittest_population), reverse=True)])            

    """
    flat = population.flatten
    return {id: size for id, size in zip(flat, np.bincount(flat)[flat])}





# pangadfs/pangadfs/pospool.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from typing import Dict, Iterable
import pandas as pd

from pangadfs.base import PospoolBase


class PospoolDefault(PospoolBase):

    def pospool(self,
                *,
                pool: pd.DataFrame,
                posfilter: Dict[str, int],
                column_mapping: Dict[str, str],
                flex_positions: Iterable[str] = ('RB', 'WR', 'TE'),
                **kwargs
                ) -> Dict[str, pd.DataFrame]:
        """Creates initial position pool
        
        Args:
            pool (pd.DataFrame):
            posfilter (Dict[str, int]): filter out low scorers by position
            column_mapping (Dict[str, str]): column names for player, position, salary, projection
            flex_positions (Iterable[str]): e.g. (WR, RB, TE)
            **kwargs: Keyword arguments for plugins (other than default)

        Returns:
            Dict[str, pd.DataFrame] where keys == posfilter.keys

        """
        d = {}
        poscol = column_mapping.get('position', 'pos')
        pointscol = column_mapping.get('points', 'proj')
        salcol = column_mapping.get('salary', 'salary')
        for position, thresh in posfilter.items():
            if position == 'FLEX':
                tmp = pool.loc[(pool[poscol].isin(flex_positions)) & (pool[pointscol] >= thresh), [pointscol, salcol]]      
            else:           
                tmp = pool.loc[(pool[poscol] == position) & (pool[pointscol] >= thresh), [pointscol, salcol]]
            prob_ = (tmp[pointscol] / tmp[salcol]) * 1000
            prob_ = prob_ / prob_.sum()
            d[position] = tmp.assign(prob=prob_)
        return d



# pangadfs/app/configapp/app.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import json
import logging
from pathlib import Path
from typing import Dict

import numpy as np
import pandas as pd
from dacite import from_dict
from stevedore.driver import DriverManager 
from stevedore.named import NamedExtensionManager

from gasettings import AppSettings
from optimizer import Optimizer


DATADIR = Path(__file__).parent.parent / 'appdata'


def main():
	"""Example application using pangadfs"""
	logging.basicConfig(level=logging.INFO)
	data = json.loads((Path(__file__).parent / 'config.json').read_text())
	ctx = from_dict(data_class=AppSettings, data=data)
	ctx.ga_settings.csvpth = DATADIR / ctx.ga_settings.csvpth

	dmgrs = {
      k: DriverManager(namespace=f'pangadfs.{k}', name=v, invoke_on_load=True)
      for k, v in ctx.plugin_settings.driver_managers.items()
	}

	emgrs = {
	  k: NamedExtensionManager(namespace=f'pangadfs.{k}', names=v, invoke_on_load=True, name_order=True)
      for k, v in ctx.plugin_settings.extension_managers.items()
	}

	# set up GeneticAlgorithm object
	opt = Optimizer(ctx=ctx, 
	                driver_managers=dmgrs, 
					extension_managers=emgrs)
	_ = opt.optimize(verbose=True)


if __name__ == '__main__':
	main()



# pangadfs/pangadfs/validate.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

import numpy as np
from numpy_indexed import unique

from pangadfs.base import ValidateBase


class DuplicatesValidate(ValidateBase):

    def validate(self,
                 *, 
                 population: np.ndarray, 
                 **kwargs) -> np.ndarray:
        """Removes duplicate individuals from population
        
            Args:
                population (np.ndarray): the population to validate
                **kwargs: keyword arguments for plugins

            Returns:
                np.ndarray: same width as population, likely has less rows

        """
        # the first part eliminates individuals with duplicate genes
        # the second part eliminates duplicate individuals
        population_sorted = np.sort(population, axis=-1)
        population = population[(population_sorted[...,1:] != population_sorted[..., :-1]).all(-1)]
        return unique(np.sort(population, axis=1))


class SalaryValidate(ValidateBase):

    def validate(self,
                 *, 
                 population: np.ndarray,
                 salaries: np.ndarray,
                 salary_cap: int, 
                 **kwargs) -> np.ndarray:
        """Ensures valid individuals in population
        
            Args:
                population (np.ndarray): the population to validate
                salaries (np.ndarray): 1D where indices are in same order as player indices
                salary_cap (int): the salary cap, e.g., 50000 or 60000
                **kwargs: keyword arguments for plugins

            Returns:
                np.ndarray: same width as population, likely has less rows

        """
        popsal = np.sum(salaries[population], axis=1)
        return population[popsal <= salary_cap]



# pangadfs/pangadfs/misc.py
# -*- coding: utf-8 -*-
# Copyright (C) 2020 Eric Truett
# Licensed under the MIT License

from typing import Iterable, Tuple

import numpy as np


def diversity(population: np.ndarray) -> np.ndarray:
    """Calculates diversity of lineups
    
    Args:
        population (np.ndarray): the population

    Returns:
        np.ndarray: is square, shape len(population) x len(population)

    """
    uniques = np.unique(population)
    a = (population[..., None] == uniques).sum(1)
    return np.einsum('ij,kj->ik', a, a)


def multidimensional_shifting(elements: Iterable, 
                              num_samples: int, 
                              sample_size: int, 
                              probs: Iterable) -> np.ndarray:
    """Based on https://medium.com/ibm-watson/incredibly-fast-random-sampling-in-python-baf154bd836a
    
    Args:
        elements (iterable): iterable to sample from, typically a dataframe index
        num_samples (int): the number of rows (e.g. initial population size)
        sample_size (int): the number of columns (e.g. team size)
        probs (iterable): is same size as elements

    Returns:
        ndarray: of shape (num_samples, sample_size)
        
    """
    replicated_probabilities = np.tile(probs, (num_samples, 1))
    random_shifts = np.random.random(replicated_probabilities.shape)
    random_shifts /= random_shifts.sum(axis=1)[:, np.newaxis]
    shifted_probabilities = random_shifts - replicated_probabilities
    samples = np.argpartition(shifted_probabilities, sample_size, axis=1)[:, :sample_size]
    return elements.to_numpy()[samples]


def parents(population: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Evenly splits population
    
    Args:
        population (np.ndarray): the population to crossover. Shape is n_individuals x n_chromosomes.

    Returns:
        Tuple[np.ndarray, np.ndarray]: population split into two equal-size arrays

    """
    fathers, mothers = np.array_split(population, 2)
    size = min(len(fathers), len(mothers))
    return fathers[:size], mothers[:size]
